#include <../arch/arch.h>

#include "winlib_common.h"
#include "winlib_win32.h"

V_LOCAL char_p LIBV_HANDLE_PROP = "LIBV_HANDLE";

/*
 * Translates wParam and lParam into scan code.
 */
V_LOCAL	unsigned
win32_make_scan_code(WPARAM wParam,LPARAM lParam)
{
	unsigned char scan = (unsigned char)((lParam>>16) & 0xFF);
	unsigned char ext = (unsigned char)((lParam & (1<<24)) != 0);
	switch(wParam)
	{
		case VK_ESCAPE:	return sc_escape;

		case VK_F1: return sc_f1;
		case VK_F2: return sc_f2;
		case VK_F3: return sc_f3;
		case VK_F4: return sc_f4;
		case VK_F5: return sc_f5;
		case VK_F6: return sc_f6;
		case VK_F7: return sc_f7;
		case VK_F8: return sc_f8;
		case VK_F9: return sc_f9;
		case VK_F10: return sc_f10;
		case VK_F11: return sc_f11;
		case VK_F12: return sc_f12;
		case VK_SCROLL: return sc_scroll_lock;

		case 192: return sc_grave;
		case '1': return sc_1;
		case '2': return sc_2;
		case '3': return sc_3;
		case '4': return sc_4;
		case '5': return sc_5;
		case '6': return sc_6;
		case '7': return sc_7;
		case '8': return sc_8;
		case '9': return sc_9;
		case '0': return sc_0;
		case 189: return sc_minus;
		case 187: return sc_equal;
		case 220: return sc_backslash;
		case VK_BACK: return sc_backspace;

		case VK_TAB: return sc_tab;
		case 'Q': return sc_q;
		case 'W': return sc_w;
		case 'E': return sc_e;
		case 'R': return sc_r;
		case 'T': return sc_t;
		case 'Y': return sc_y;
		case 'U': return sc_u;
		case 'I': return sc_i;
		case 'O': return sc_o;
		case 'P': return sc_p;
		case 219: return sc_lbracket;
		case 221: return sc_rbracket;
		case VK_RETURN:	return ext ? sc_kp_enter : sc_enter;

		case VK_CAPITAL: return sc_caps_lock;
		case 'A': return sc_a;
		case 'S': return sc_s;
		case 'D': return sc_d;
		case 'F': return sc_f;
		case 'G': return sc_g;
		case 'H': return sc_h;
		case 'J': return sc_j;
		case 'K': return sc_k;
		case 'L': return sc_l;
		case 186: return sc_semicolon;
		case 222: return sc_apostrophe;

		case 16: return (scan == 42) ? sc_lshift : sc_rshift;
		case 'Z': return sc_z;
		case 'X': return sc_x;
		case 'C': return sc_c;
		case 'V': return sc_v;
		case 'B': return sc_b;
		case 'N': return sc_n;
		case 'M': return sc_m;
		case 188: return sc_comma;
		case 190: return sc_period;
		case 191: return sc_slash;

		case 17: return ext ? sc_rcontrol : sc_lcontrol;
		case VK_LWIN: return sc_lsuper;
		case 18: return ext ? sc_ralt : sc_lalt;
		case VK_SPACE: return sc_space;
		case VK_RWIN: return sc_rsuper;
		case VK_APPS: return sc_menu;
		case VK_NUMLOCK: return sc_num_lock;
		case VK_DIVIDE: return sc_kp_slash;
		case 106: return sc_kp_multiply;
		case 109: return sc_kp_subtract;
		case 107: return sc_kp_add;
		case VK_CLEAR: return sc_kp_begin;
	}
	switch(scan)
	{
		case 71: return ext ? sc_home : sc_kp_home;
		case 72: return ext ? sc_up : sc_kp_up;
		case 73: return ext ? sc_prior : sc_kp_prior;
		case 75: return ext ? sc_left : sc_kp_left;
		case 77: return ext ? sc_right : sc_kp_right;
		case 79: return ext ? sc_end : sc_kp_end;
		case 80: return ext ? sc_down : sc_kp_down;
		case 81: return ext ? sc_next : sc_kp_next;
		case 82: return ext ? sc_insert : sc_kp_insert;
		case 83: return ext ? sc_delete : sc_kp_delete;
		case 69: return sc_pause;
		case 55: return sc_print;
	}
	return sc_unknown;
}

V_LOCAL void
win32_get_key_modifiers(ui_event_p ev)
{
	u8_t keys[256];
	memZero(keys,256);
	GetKeyboardState(keys);
	ev->state = ev->button = 0;
	if(keys[VK_NUMLOCK] & 1)
		ev->state |= UI_NUM_LOCK_MASK;
	if(keys[VK_CAPITAL] & 1)
		ev->state |= UI_CAPS_LOCK_MASK;
	if(keys[VK_SCROLL] & 1)
		ev->state |= UI_SCROLL_LOCK_MASK;
    if(keys[VK_LCONTROL] & 0x80 || keys[VK_RCONTROL] & 0x80)
		ev->state |= UI_CONTROL_MASK;
    if(keys[VK_LSHIFT] & 0x80 || keys[VK_RSHIFT] & 0x80)
		ev->state |= UI_SHIFT_MASK;
	if(keys[18] & 0x80)
		ev->state |= UI_ALT_MASK;
    if(keys[VK_LMENU] & 0x80 || keys[VK_RMENU] & 0x80)
		ev->state |= UI_SUPER_MASK;
    if(keys[VK_LBUTTON] & 0x80)
		ev->button |= UI_LBUTTON_MASK;
    if(keys[VK_MBUTTON] & 0x80)
		ev->button |= UI_MBUTTON_MASK;
    if(keys[VK_RBUTTON] & 0x80)
		ev->button |= UI_RBUTTON_MASK;
}

V_LOCAL void
win32_get_mouse_pos(HWND wnd,POINT *psc,ui_event_p ev)
{
	POINT pcc;
	pcc.x = ev->screen_x = psc->x;
	pcc.y = ev->screen_y = psc->y;
	if(ScreenToClient(wnd,&pcc))
	{
		ev->client_x = pcc.x;
		ev->client_y = pcc.y;
	}
	else
	{
		ev->client_x = psc->x;
		ev->client_y = psc->y;
	}
}

V_LOCAL void
win32_get_char_codes(HWND wnd,WPARAM wParam,LPARAM lParam,ui_event_p ev)
{
	MSG msg, result;
	/* flush all pending WM_CHARs from queue */
	while(PeekMessage(&msg,wnd,WM_CHAR,WM_CHAR,PM_REMOVE));
	/* make fake WM_KEYDOWN message */
	msg.hwnd = wnd;
	msg.message = WM_KEYDOWN;
	msg.wParam = wParam;
	msg.lParam = lParam;
	msg.pt.x = msg.pt.y = msg.time = 0;
	ev->key.ch = ev->key.unicode = 0;
	/* translate it to ASCII */
	TranslateMessage(&msg);
	if(!PeekMessageA(&result,wnd,WM_CHAR,WM_CHAR,PM_REMOVE))
		return;
	ev->key.ch = result.wParam;
	/* translate it to UNICODE */
	TranslateMessage(&msg);
	if(!PeekMessageW(&result,wnd,WM_CHAR,WM_CHAR,PM_REMOVE))
		return;
	ev->key.unicode = result.wParam;
}

#define	SIGNED_LOWORD(src) \
	(((src) & 0x8000) ? LOWORD(src) | 0xFFFF0000 : LOWORD(src))

#define	SIGNED_HIWORD(src) \
	(((src) & 0x80000000) ? HIWORD(src) | 0xFFFF0000 : HIWORD(src))

V_LOCAL void
win32_window_proc(arch_window_p aw,
	UINT msg,WPARAM wParam,LPARAM lParam,DWORD time,POINT *psc)
{
	POINT def_psc;
	RECT rc;
	PAINTSTRUCT ps;
	ui_event_t dst;
	vboolean zap_message = false;
	if(!time)
		time = GetTickCount();
	if(!psc)
	{
		psc = &def_psc;
		GetCursorPos(psc);
	}
	dst.time = ui_make_message_time(&aw->start_time,time);
	dst.type = ui_no_event;
	switch(msg)
	{
		case WM_ACTIVATE:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
			if((msg == WM_ACTIVATE && (LOWORD(wParam) != WA_INACTIVE)) ||
				msg == WM_SETFOCUS)
			{
				aw->focused = true;
				dst.type = ui_window_focus_in_event;
			}
			else
			{
				aw->focused = false;
				dst.type = ui_window_focus_out_event;
			}
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_CLOSE:
			dst.type = ui_wm_event;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.configure.wm_event = ui_wm_close;
			break;

		case WM_PAINT:
			memZero(&rc, sizeof(RECT));
			if(!GetUpdateRect(aw->wnd,&rc,FALSE))
				return;
			BeginPaint(aw->wnd,&ps);
			EndPaint(aw->wnd,&ps);
			dst.type = ui_window_expose_event;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.configure.exposed_left = rc.left;
			dst.configure.exposed_top = rc.top;
			dst.configure.exposed_right = rc.right;
			dst.configure.exposed_bottom = rc.bottom;
			break;

		case WM_MOVE:
			if(GetUpdateRect(aw->wnd,&rc,FALSE))
				PatBlt(aw->dc,rc.left,rc.top,
					rc.right - rc.left,rc.bottom-rc.top,BLACKNESS);
			dst.type = ui_window_move_event;
			aw->x = (int)LOWORD(lParam);
			aw->y = (int)HIWORD(lParam);
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_SIZE:
			if(GetUpdateRect(aw->wnd,&rc,FALSE))
				PatBlt(aw->dc,rc.left,rc.top,
					rc.right - rc.left,rc.bottom-rc.top,BLACKNESS);
			dst.type = ui_window_size_event;
			aw->width = (unsigned)LOWORD(lParam);
			aw->height = (unsigned)HIWORD(lParam);
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYUP:
			if(msg == WM_KEYDOWN || msg == WM_SYSKEYDOWN)
				dst.key.press = true;
			else
				dst.key.press = false;
#ifndef	CONFIG_ENABLE_AUTOREPEAT
			if(dst->key.press && (lParam & (1 << 30)))
			{
				zap_message = true;
				break;
			}
#endif	/* !CONFIG_ENABLE_AUTOREPEAT */
			dst.type = ui_key_event;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			dst.key.scan = win32_make_scan_code(wParam,lParam);
			win32_get_char_codes(aw->wnd,wParam,lParam,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			if(dst.key.scan != sc_unknown)
				aw->keyboard_state[dst.key.scan] = dst.key.press ? 1 : 0;
			break;

		case WM_MOUSEMOVE:
			dst.type = ui_pointer_motion_event;
			ui_make_default_key_event(&dst.key);
			dst.client_x = SIGNED_LOWORD(lParam);
			dst.client_y = SIGNED_LOWORD(lParam);
			dst.screen_x = psc->x;
			dst.screen_y = psc->y;
			win32_get_key_modifiers(&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MBUTTONUP:
		case WM_RBUTTONUP:
			dst.type = ui_key_event;
			if(msg == WM_LBUTTONDOWN || msg == WM_MBUTTONDOWN ||
					msg == WM_RBUTTONDOWN)
				dst.key.press = true;
			else
				dst.key.press = false;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(&dst);
			dst.key.ch = dst.key.unicode = 0; 
			if(msg == WM_LBUTTONDOWN || msg == WM_LBUTTONUP)
				dst.key.scan = sc_lbutton;
			else if(msg == WM_RBUTTONDOWN || msg == WM_RBUTTONUP)
				dst.key.scan = sc_rbutton;
			else
				dst.key.scan = sc_mbutton;
			aw->keyboard_state[dst.key.scan] = dst.key.press ? 1 : 0;
			break;

		default:
			zap_message = true;
			break;
	}
	if(!zap_message)
		aw->msg_que = eventDeqPush(aw->msg_que,&dst);
	return;
}

/*
 * Default callback. All system-generated messages and messages
 * unknown to win32_window_proc() have to be processed here.
 */
V_LOCAL LRESULT CALLBACK
win32_sys_window_proc(HWND wnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	arch_window_p aw = (arch_window_p)GetProp(wnd,LIBV_HANDLE_PROP);
	if(!aw || aw->wnd != wnd)
		aw = NULL;
	switch(msg)
	{
		case WM_CREATE:
			return 0;

		case WM_PAINT:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			return 0;

		case WM_ERASEBKGND:
			return 1; /* no further erasing required */
		/*
		 * These are nonqueued messages, so we have to transfer control by hand.
		 */
		case WM_ACTIVATE:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			break;

		case WM_SETFOCUS:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			break;

		case WM_KILLFOCUS:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			break;

		case WM_MOVE:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			break;

		case WM_SIZE:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			break;

		case WM_CLOSE:
			if(aw)
				win32_window_proc(aw,msg,wparam,lparam,0,NULL);
			break;

		default:
			break;
	}
	return DefWindowProc(wnd,msg,wparam,lparam);
}

V_LOCAL void
win32_query_vid_modes(arch_window_p aw)
{
	DWORD num;
	DEVMODE dm;
	fifo_t modes;
	memZero(&dm, sizeof(DEVMODE));
	dm.dmSize = sizeof(dm);
	fifoInit(&modes, sizeof(DEVMODE));
	for(num = 0;EnumDisplaySettings(NULL, num, &dm);num++)
	{
		if(dm.dmBitsPerPel >= 15) /* Hi/TrueColor */
			fifoPush(&modes,&dm);
	}
	if(aw->vidmodes)
	{
		memFree(aw->vidmodes);
		aw->vidmodes = NULL;
	}
	aw->num_vidmodes = 0;
	if(num)
	{
		aw->vidmodes = (vidmode_p)memAllocZ(num * sizeof(vidmode_t));
		if(aw->vidmodes)
		{
			vidmode_p ptr = aw->vidmodes;
			while(fifoPop(&modes,&dm))
			{
				ptr->width = dm.dmPelsWidth;
				ptr->height = dm.dmPelsHeight;
				ptr->bpp = dm.dmBitsPerPel;
				ptr->refresh = dm.dmDisplayFrequency;
				if(ptr->refresh == 1)
					ptr->refresh = 0;
				ptr->modeline = aw->num_vidmodes;
				aw->num_vidmodes++;
				ptr++;
			}
		}
	}
	fifoClear(&modes);
}

V_LOCAL LONG
win32_change_video_mode_1(vidmode_p new_mode,
	u32_t old_width,u32_t old_height,u32_t old_bpp,u32_t old_refresh)
{
	ARCH_CALLER_NAME(win32_change_video_mode)
	if(new_mode)
	{
		DEVMODE dm;
		memZero(&dm, sizeof(DEVMODE));
		dm.dmSize = sizeof(dm);
		if(!new_mode->width)
		{
			ARCH_INFO("Using current desktop width");
			dm.dmPelsWidth = old_width;
		}
		else
			dm.dmPelsWidth = new_mode->width;
		if(!new_mode->height)
		{
			ARCH_INFO("Using current desktop height");
			dm.dmPelsHeight = old_height;
		}
		else
			dm.dmPelsHeight = new_mode->height;
		if(!new_mode->bpp)
		{
			ARCH_INFO("Using current desktop display depth");
			dm.dmBitsPerPel = old_bpp;
		}
		else
			dm.dmBitsPerPel = new_mode->bpp;
		dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
		if(ntGetVersion() > 0)
		{
			if(!new_mode->refresh)
			{
				if(dm.dmPelsWidth > old_width || dm.dmPelsHeight > old_height)
				{
					ARCH_WARNING("No refresh rate for hi-res mode; using 75 Hz");
					/* suitable both for CRT and (good) LCD monitors */
					dm.dmDisplayFrequency = 75;
				}
				else
				{
					ARCH_INFO("Using current refresh rate");
					dm.dmDisplayFrequency = old_refresh;
				}
			}
			else
				dm.dmDisplayFrequency = new_mode->refresh;
			dm.dmFields |= DM_DISPLAYFREQUENCY;
		}
		else
		{
			if(new_mode->refresh)
				ARCH_WARNING("Unable to change refresh rate under Windows 9x");
		}
		return ChangeDisplaySettings(&dm,0);
	}
	else
		return ChangeDisplaySettings(NULL,0);
}

V_LOCAL vboolean
win32_change_video_mode_2(vidmode_p mode)
{
	ARCH_CALLER_NAME(archChangeVideoMode)
	HDC refdc = GetDC(NULL);
	LONG retval;
	u32_t bpp = GetDeviceCaps(refdc,BITSPIXEL);
	u32_t refresh = GetDeviceCaps(refdc,VREFRESH);
	u32_t width = GetSystemMetrics(SM_CXSCREEN);
	u32_t height = GetSystemMetrics(SM_CYSCREEN);
	ReleaseDC(NULL,refdc);
	retval = win32_change_video_mode_1(mode,width,height,bpp,refresh);
	if(retval == -1)
	{
		/*
		 * Try to switch away from fullscreen text mode.
		 */
		arch_window_p dummy;
		dummy = archWinOpen("dummy",0,0,320,200,false,false);
		if(dummy)
		{
			archWinShow(dummy,true);
			archWinFocus(dummy);
			timeDelay(500);
			retval = win32_change_video_mode_1(mode,width,height,bpp,refresh);
			archWinClose(dummy);
		}
	}
	if(retval != DISP_CHANGE_SUCCESSFUL)
	{
		ARCH_ERROR("Unable to change video mode");
		return false;
	}
	return true;
}

vboolean V_CALL
archWinSetVidMode(arch_window_p aw,int new_mode)
{
	vidmode_p mode;
	if(new_mode == -1)
	{
		/* restore default mode */
		if(aw->current_vidmode != -1)
		{
			win32_change_video_mode_2(NULL);
			SetWindowPos(aw->wnd,HWND_NOTOPMOST,0,0,0,0,
				SWP_NOSIZE|SWP_NOMOVE);
			aw->current_vidmode = -1;
		}
		return true;
	}
	if(new_mode >= aw->num_vidmodes)
		return false;
	mode = aw->vidmodes + new_mode;
	if(win32_change_video_mode_2(mode))
	{
		aw->current_vidmode = new_mode;
		SetWindowPos(aw->wnd,HWND_TOP,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE);
		return true;
	}
	return false;
}

V_LOCAL vboolean
win32_register_window_class(arch_window_p aw)
{
	ARCH_CALLER_NAME(win32_register_window_class)
	strCopy(aw->class_name,"WC_CLASS_",MAX_CLASS_NAME);
	archUniqueString(aw->class_name+9,MAX_CLASS_NAME-9);
	memZero(&aw->wc,sizeof(WNDCLASS));
	aw->wc.lpszClassName = aw->class_name;
	aw->wc.hInstance = GetModuleHandle(NULL);
	aw->wc.style = CS_CLASSDC;
	aw->wc.lpfnWndProc = win32_sys_window_proc;
	aw->wc.hIcon = LoadIcon(NULL,IDI_WINLOGO);
	aw->wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	aw->wc.hbrBackground = GetStockObject(HOLLOW_BRUSH);
	aw->wc.lpszMenuName = NULL;
	if(RegisterClass(&aw->wc))
		return true;
	SYSCALL_FAILED("RegisterClass");
	return false;
}

arch_window_p V_CALL
archWinOpenNP(void_p osd_parent,
	char *title,s32_t left,s32_t top,u32_t width,u32_t height,
	vboolean caption,vboolean allow_resize)
{
	ARCH_CALLER_NAME(archWinOpenNP)
	RECT rc;
	arch_window_p aw;
	DWORD style = WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
	if(allow_resize)
	{
		if(!caption)
		{
			ARCH_ERROR("can't create resizeable window without caption");
			return NULL;
		}
		style |= WS_OVERLAPPEDWINDOW;
	}
	else
	{
		style |= WS_POPUP;
		if(caption)
			style |= WS_CAPTION;
	}
	aw = (arch_window_p)memAllocZ(sizeof(arch_window_t));
	if(!aw)
	{
		ARCH_ERROR("out of memory");
		return NULL;
	}
	INIT_COMMON_WINDOW_FIELDS(aw);
	if(!win32_register_window_class(aw))
	{
		archWinClose(aw);
		return NULL;
	}
	if(title)
		strCopy(aw->title,title,MAX_WINDOW_TITLE);
	aw->start_time = GetTickCount();
	aw->wnd = CreateWindow(aw->class_name,
		aw->title, style, left, top, width, height,
		osd_parent,
		NULL,
		GetModuleHandle(NULL),NULL);
	if(!aw->wnd)
	{
		SYSCALL_FAILED("CreateWindow");
		archWinClose(aw);
		return NULL;
	}
	SetProp(aw->wnd,LIBV_HANDLE_PROP,(HANDLE)aw);
	GetClientRect(aw->wnd,&rc);
	aw->delta_w = width - rc.right;
	aw->delta_h = height - rc.bottom;
	archWinSize(aw,width,height);
	aw->dc = GetDC(aw->wnd);
	win32_query_vid_modes(aw);
	return aw;
}

arch_window_p V_CALL
archWinOpen(char *title,s32_t left,s32_t top,u32_t width,u32_t height,
	vboolean caption,vboolean allow_resize)
{
	return archWinOpenNP(NULL,title,left,top,width,height,
		caption,allow_resize);
}

vboolean	V_CALL
archWinShow(arch_window_p aw,vboolean show)
{
	if(show)
	{
		if(!aw->visible)
		{
			aw->visible = ShowWindow(aw->wnd,SW_SHOW) ? true : false;
			UpdateWindow(aw->wnd);
		}
		return aw->visible;
	}
	aw->visible = ShowWindow(aw->wnd,SW_HIDE) ? false : true;
	return !aw->visible;
}

void V_CALL
archWinFocus(arch_window_p aw)
{
	SetFocus(aw->wnd);
	SetActiveWindow(aw->wnd);
}

vboolean	V_CALL
archWinMove(arch_window_p aw,s32_t left,s32_t top)
{
	return SetWindowPos(aw->wnd,HWND_NOTOPMOST,left,top,0,0,
		SWP_NOSIZE | SWP_NOZORDER) ? true : false;
}

vboolean	V_CALL
archWinSize(arch_window_p aw,s32_t width,s32_t height)
{
	return SetWindowPos(aw->wnd,HWND_NOTOPMOST,
		0,0,width + aw->delta_w,height + aw->delta_h,
		SWP_NOMOVE | SWP_NOZORDER) ? true : false;
}

vboolean V_CALL
archWinShowPointer(arch_window_p aw,vboolean show)
{
	vboolean result;
	if(show)
	{
		if(!--aw->pointer_hide_count)
			result = (ShowCursor(TRUE) < 0) ? true : false;
	}
	else
	{
		if(!aw->pointer_hide_count++)
			result = (ShowCursor(FALSE) < 0) ? true : false;
	}
	return result;
}

vboolean	V_CALL
archWinWarpPointer(arch_window_p aw,s32_t x,s32_t y)
{
	POINT pt;
	pt.x = x;
	pt.y = y;
	ClientToScreen(aw->wnd,&pt);
	return SetCursorPos(pt.x,pt.y) ? true : false;
}

void V_CALL
archWinCapturePointer(arch_window_p aw,vboolean capture)
{
	if(capture)
	{
		if(aw->pointer_captured)
			return;
		SetCapture(aw->wnd);
		aw->pointer_captured = true;
	}
	else
	{
		if(!aw->pointer_captured)
			return;
		ReleaseCapture();
		aw->pointer_captured = false;
	}
}

void	V_CALL
archWinClose(arch_window_p aw)
{
	if(aw->pointer_hide_count > 0)
		ShowCursor(TRUE);
	/*
	 * Restore original video mode
	 */
	if(aw->current_vidmode != -1)
		archWinSetVidMode(aw,-1);
	if(aw->wnd)
	{
		if(aw->dc)
			ReleaseDC(aw->wnd,aw->dc);
		RemoveProp(aw->wnd,LIBV_HANDLE_PROP);
		DestroyWindow(aw->wnd);
	}
	UnregisterClass(aw->class_name,GetModuleHandle(NULL));
	CLEAR_COMMON_WINDOW_FIELDS(aw);
	memFree(aw);
}

V_LOCAL void
win32_poll_nb(arch_window_p aw)
{
	MSG msg;
	while(PeekMessage(&msg,aw->wnd,0,0,PM_REMOVE))
		win32_window_proc(aw,msg.message,msg.wParam,msg.lParam,
			msg.time, &msg.pt);
}

void V_CALL
archWinPoll(arch_window_p aw,vboolean block)
{
	if(!block)
		win32_poll_nb(aw);
	else
	{
		while(!aw->msg_que)
		{
			WaitMessage();
			win32_poll_nb(aw);
		}
	}
}

/* ====================================================================================== */

#include "unitabs.c"

#include "winlib_common.c"

void dump_event(ui_event_p ev)
{
	charmap_char_description_p cd;
	printf("=============================\nType: ");
	switch(ev->type)
	{
		case ui_key_event: puts("ui_key_event"); break;
		case ui_pointer_motion_event: puts("ui_pointer_motion_event"); break;
		case ui_window_focus_in_event: puts("ui_window_focus_in_event"); break;
		case ui_window_focus_out_event: puts("ui_window_focus_out_event"); break;
		case ui_window_expose_event: puts("ui_window_expose_event"); break;
		case ui_window_map_event: puts("ui_window_map_event"); break;
		case ui_window_unmap_event: puts("ui_window_unmap_event"); break;
		case ui_window_move_event: puts("ui_window_move_event"); break;
		case ui_window_size_event: puts("ui_window_size_event"); break;
		default:
			puts("unknown");
			return;
	}
	printf("Time %g, x %d, y %d, rx %d, ry %d\n",
		ev->time, ev->client_x, ev->client_y, ev->screen_x, ev->screen_y);
	printf("Buton %x, state %x\n",ev->button,ev->state);
	printf("Press %u, char '%c', uni %x, scan %s\n",
		ev->key.press, (ev->key.ch >= 32) ? ev->key.ch : '?',
		ev->key.unicode,
		uiKeyscan2Name(ev->key.scan));
	if(ev->type == ui_key_event && ev->key.unicode)
	{
		cd = find_char_desc_by_uni(latin1_ccd,ev->key.unicode);
		if(!cd)
			cd = find_char_desc_by_uni(koi8_r_ccd,ev->key.unicode);
		if(cd)
			printf("\t%s\n",cd->char_name);
	}
	printf("Focused %u, visible %u\n",
		ev->configure.focused, ev->configure.visible);
	printf("Exposed: [%u,%u]-[%u,%u]\n",
		ev->configure.exposed_left,
		ev->configure.exposed_top,
		ev->configure.exposed_right,
		ev->configure.exposed_bottom);
	printf("Position [%d,%d]\n",
		ev->configure.x, ev->configure.y);
	printf("Size %u X %u\n",
		ev->configure.width, ev->configure.height);
}

int main(void)
{
	arch_window_p w;
	ui_event_t ev;

	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);

	w = archWinOpen("test",100,100,640,480,true,true);
	archWinShow(w,true);
	
	for(;;)
	{
		archWinPoll(w,true);
		w->msg_que = eventDeqPopBack(w->msg_que, &ev);
		dump_event(&ev);
		if(ev.type == ui_key_event && ev.key.scan == sc_escape)
			break;
	}

	archWinClose(w);

	memPrintStats();
	return 0;
}
